package mdx

  1. Overview
  2. Docs

Module Mdx_topSource

Toplevel logic for mdx.

Sourcetype t

The type for configuration values.

Sourcetype directive =
  1. | Directory of string
  2. | Load of string
    (*

    The type for toplevel directives

    *)
Sourceval init : verbose:bool -> silent:bool -> verbose_findlib:bool -> directives:directive list -> packages:string list -> predicates:string list -> unit -> t

init () is a new configuration value.

Sourceval eval : t -> string list -> (string list, string list) result

eval t p evaluates the toplevel phrase p (possibly spawning on mulitple lines) with the configuration value t.

Sourceval in_env : Mdx.Ocaml_env.t -> (unit -> 'a) -> 'a
Sourceval skip_reason : t -> string option

skip_reason t returns the reason passed to mdx_skip, or None if that hasn't been called yet.

Sourceval clear_skip : t -> unit

clear_skip t clears the skip state, so that skip_reason t returns None again.