package mdx

  1. Overview
  2. Docs

Module Mdx.LibrarySource

Sourcetype t = {
  1. base_name : string;
  2. sub_lib : string option;
}

The type to represent dune libraries as referred to in '#require' statements or in dune files. I.e. lib.sub_lib is {base_name = "lib"; sub_lib = Some "sub_lib"}.

Sourceval equal : t -> t -> bool
Sourceval compare : t -> t -> int
Sourceval pp : t Fmt.t
Sourceval from_string : string -> (t, string) Result.result

from_string s returns the library represented by s or an error if s isn't a valid library.

Sourcemodule Set : sig ... end

Set of libraries